from relevanceai.base import Base
from relevanceai.vector_tools.cluster import Cluster
from relevanceai.vector_tools.dim_reduction import DimReduction

class VectorTools(Base):
    """Vector Tools Client"""
    def __init__(self, project: str, api_key: str):
        self.cluster = Cluster(project=project, api_key=api_key)
        self.dim_reduction = DimReduction(project=project, api_key=api_key)
        super().__init__(project, api_key) 



"""Batch Retrieve"""

from typing import List
from relevanceai.api.endpoints.client import APIClient
from relevanceai.api.batch.chunk import Chunker

BYTE_TO_MB = 1024 * 1024
LIST_SIZE_MULTIPLIER = 3


class BatchRetrieve(APIClient, Chunker):
    def get_documents(
        self,
        dataset_id: str,
        filters: list = [],
        number_of_documents: int = 20,
        cursor: str = None,
        batch_size: int = 1000,
        sort: list = [],
        select_fields: list = [],
        include_vector: bool = True):
        
        """
        Retrieve documents with filters. Filter is used to retrieve documents that match the conditions set in a filter query. This is used in advance search to filter the documents that are searched. \n
        If you are looking to combine your filters with multiple ORs, simply add the following inside the query {"strict":"must_or"}.
        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        number_of_documents: int
            Number of documents to retrieve
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        cursor: string
            Cursor to paginate the document retrieval
        batch_size: int
            Number of documents to retrieve per iteration
        include_vector: bool
            Include vectors in the search results
        sort: list
            Fields to sort by. For each field, sort by descending or ascending. If you are using descending by datetime, it will get the most recent ones.
        filters: list
            Query for filtering the search results
        """
        if batch_size > number_of_documents:
            batch_size = number_of_documents 

        resp = self.datasets.documents.get_where(
            dataset_id=dataset_id,
            select_fields=select_fields,
            include_vector=include_vector,
            page_size=batch_size,
            sort = sort,
            is_random=False,
            random_state=0,
            filters=filters,
            cursor=cursor
        )
        data = resp["documents"]

        if number_of_documents > batch_size:
            _cursor = resp["cursor"]
            _page = 0
            while resp:
                self.logger.debug(f"Paginating {_page} batch size {batch_size} ...")
                resp = self.datasets.documents.get_where(
                    dataset_id=dataset_id,
                    select_fields=select_fields,
                    include_vector=include_vector,
                    page_size=batch_size,
                    sort = sort,
                    is_random=False,
                    random_state=0,
                    filters=filters,
                    cursor=_cursor
                )
                _data = resp["documents"]
                _cursor = resp["cursor"]
                if (_data == []) or (_cursor == []):
                    break
                data += _data
                if number_of_documents and (len(data) >= int(number_of_documents)):
                    break
                _page += 1
            data = data[:number_of_documents]

        return data
    
    def get_all_documents(
        self,
        dataset_id: str,
        chunk_size: int = 10000,
        filters: List = [],
        sort: List = [],
        select_fields: List = [],
        include_vector: bool = True
    ):
        """
        Retrieve all documents with filters. Filter is used to retrieve documents that match the conditions set in a filter query. This is used in advance search to filter the documents that are searched. For more details see documents.get_where.

        Example
        ---------

        >>> client = Client()
        >>> client.get_all_documents(dataset_id="sample_dataset"")

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        chunk_size : list
            Number of documents to retrieve per retrieval
        include_vector: bool
            Include vectors in the search results
        sort: list
            Fields to sort by. For each field, sort by descending or ascending. If you are using descending by datetime, it will get the most recent ones.
        filters: list
            Query for filtering the search results
        select_fields : list
            Fields to include in the search results, empty array/list means all fields.
        """

        # Initialise values
        length = 1
        cursor = None
        full_data = []

        # While there is still data to fetch, fetch it at the latest cursor
        while length > 0:
            x = self.datasets.documents.get_where(
                dataset_id,
                filters=filters,
                cursor=cursor,
                page_size=chunk_size,
                sort=sort,
                select_fields=select_fields,
                include_vector=include_vector
            )
            length = len(x["documents"])
            cursor = x["cursor"]

            # Append fetched data to the full data
            if length > 0:
                full_data += x["documents"]
        return full_data


"""Chunk Helper functions"""
from typing import List, Union

import pandas as pd

from relevanceai.progress_bar import progress_bar

class Chunker:
    """Update the chunk Mixins"""
    def chunk(self, documents: Union[pd.DataFrame, List], chunksize: int=20):
        """
        Chunk an iterable object in Python. \n
        Example:

        >>> documents = [{...}]
        >>> ViClient.chunk(documents)

        Parameters
        ----------
        documents:
            List of dictionaries/Pandas dataframe
        chunksize:
            The chunk size of an object.
        """
        if isinstance(documents, pd.DataFrame):
            for i in progress_bar(range(0, int(len(documents) / chunksize))):
                yield documents.iloc[i * chunksize : (i + 1) * chunksize]
        else:
            for i in progress_bar(range(0, int(len(documents) / chunksize))):
                yield documents[i * chunksize : ((i + 1) * chunksize)]



"""Prediction services
"""
from relevanceai.base import Base

class Prediction(Base):
    def __init__(self, project, api_key):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)

    def KNN(
        self,
        dataset_id: str,
        vector: list,
        vector_field: str,
        target_field: str,
        k: int = 5,
        weighting: bool or list = True,
        impute_value: int = 0,
        predict_operation: str = "most_frequent",
        include_search_results: bool = True,
    ):

        """ 
        Predict using KNN regression.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        vector: list
            Vector, a list/array of floats that represents a piece of data.
        vector_field: string
            The vector field to search in. It can either be an array of strings (automatically equally weighted) (e.g. ['check_vector_', 'yellow_vector_']) or it is a dictionary mapping field to float where the weighting is explicitly specified (e.g. {'check_vector_': 0.2, 'yellow_vector_': 0.5})
        target_field: string
            The field to perform regression on.
        k: int
            The number of results for KNN.
        weighting : bool/list
            The weighting for each prediction
        impute_value: int
            The value used to fill in the document when the data is missing.
        predict_operation: string
            How to predict using the vectors. One of most_frequent or `sum_scores
        include_search_results: bool
            Whether to include search results.
        """

        return self.make_http_request(
            f"/services/prediction/regression/knn",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "vector": vector,
                "vector_field": vector_field,
                "target_field": target_field,
                "k": k,
                "weighting": weighting,
                "impute_value": impute_value,
                "predict_operation": predict_operation,
                "include_search_results": include_search_results
            }
        )

    def KNN_from_results(
        self,
        field: str,
        results: list,
        impute_value: int = 0,
        predict_operation: str = "most_frequent"
    ):

        """ 
        Predict using KNN regression from search results

        Parameters
        ----------
        field: string
            Field in results to use for the prediction. Can be multiplied with weighting.
        results: dict
            List of results in a dictionary
        weighting : bool/list
            The weighting for each prediction
        impute_value: int
            The value used to fill in the document when the data is missing.
        predict_operation: string
            How to predict using the vectors. One of most_frequent or `sum_scores
        """

        return self.make_http_request(
            f"/services/prediction/regression/knn_from_results",
            method="POST",
            parameters={
                "field": field,
                "results": results,
                "impute_value": impute_value,
                "predict_operation": predict_operation
            }
        )





# -*- coding: utf-8 -*-

from abc import abstractmethod
import numpy as np
import warnings

from typing import List, Union, Dict, Any, Optional

from doc_utils import DocUtils

from relevanceai.api.client import BatchAPIClient
from relevanceai.logger import LoguruLogger
from relevanceai.vector_tools.constants import CLUSTER, CLUSTER_DEFAULT_ARGS
from relevanceai.errors import ClusteringResultsAlreadyExistsError

class ClusterBase(LoguruLogger, DocUtils):
    def __call__(self, *args, **kwargs):
        return self.fit_transform(*args, **kwargs)

    @abstractmethod
    def fit_transform(self, vectors):
        """
        """
        raise NotImplementedError
    
    def fit_documents(
        self,
        vector_field: list,
        docs: list,
        alias: str="default",
        cluster_field: str="_cluster_",
        return_only_clusters: bool=True
    ):
        """
        Train clustering algorithm on documents and then store the labels
        inside the documents.
        
        Parameters
        -----------
        vector_field: list
            The vector field of the documents
        docs: list
            List of documents to run clustering on
        alias: str
            What the clusters can be called
        cluster_field: str
            What the cluster fields should be called
        return_only_clusters: bool
            If True, return only clusters, otherwise returns the original document

        """
        if len(vector_field) == 1:
            # filtering out entries not containing the specified vector
            docs = list(filter(DocUtils.list_doc_fields, docs))
            vectors = self.get_field_across_documents(vector_field[0], docs)
        else:
            raise ValueError("We currently do not support more than 1 vector field yet. This will be supported in the future.")
        cluster_labels = self.fit_transform(vectors)
        # Label the clusters
        cluster_labels = self._label_clusters(cluster_labels)
        self.set_field_across_documents(
            f"{cluster_field}.{vector_field[0]}.{alias}", cluster_labels, docs
        )
        if return_only_clusters:
            return [{"_id": d.get("_id"), cluster_field: d.get(cluster_field)} for d in docs]
        return docs

    def to_metadata(self):
        """You can also store the metadata of this clustering algorithm
        """
        raise NotImplementedError
    
    def _label_cluster(self, label: Union[int, str]):
        if isinstance(label, (int, float)):
            return "cluster_" + str(label)
        return str(label)

    def _label_clusters(self, labels):
        return [self._label_cluster(x) for x in labels]

class CentroidCluster(ClusterBase):
    def __call__(self, *args, **kwargs):
        return self.fit_transform(*args, **kwargs)

    @abstractmethod
    def fit_transform(self, vectors):
        raise NotImplementedError
    
    @abstractmethod
    def get_centers(self) -> Union[np.ndarray, List[list]]:
        """Get centers for the centroid-based clusters
        """
        raise NotImplementedError
    
    def get_centroid_docs(self) -> List:
        """Get the centroid documents to store.
        """
        self.centers = self.get_centers()
        if isinstance(self.centers, np.ndarray):
            self.centers = self.centers.tolist()
        return [
            {
                "_id": f"cluster_{i}",
                "centroid_vector_": self.centers[i]
            } for i in range(len(self.centers))
        ]

class DensityCluster(ClusterBase):
    def __call__(self, *args, **kwargs):
        return self.fit_transform(*args, **kwargs)

    def fit_transform(self, vectors):
        raise NotImplementedError


class MiniBatchKMeans(CentroidCluster):
    def __init__(
        self,
        k: Union[None, int] = 10,
        init: str = "k-means++",
        verbose: bool = True,
        compute_labels: bool = True,
        max_no_improvement: int=2
     ):
        """
        Kmeans Centroid Clustering

        Parameters
        ------------
        k: int
            The number of clusters
        init: str
            The optional parameter to be clustering
        verbose: bool
            If True, will print what is happening 
        compute_labels: bool
            If True, computes the labels of the cluster 
        max_no_improvement: int
            The maximum number of improvemnets
        """
        self.k = k
        self.init = init
        self.verbose = verbose
        self.compute_labels = compute_labels
        self.max_no_improvement = max_no_improvement

    def _init_model(self):
        from sklearn.cluster import MiniBatchKMeans
        self.km = MiniBatchKMeans(
            n_clusters=self.k, 
            init=self.init,
            verbose=self.verbose,
            compute_labels=self.compute_labels,
            max_no_improvement=self.max_no_improvement
        )
        return

    def fit_transform(
        self,
        vectors: Union[np.ndarray, List]
    ):
        """
        Fit and transform transform the vectors
        """
        if not hasattr(self, "km"):
            self._init_model()
        self.km.fit(vectors)
        cluster_labels = self.km.labels_.tolist()
        # cluster_centroids = km.cluster_centers_
        return cluster_labels

    def get_centers(self):
        """Returns centroids of clusters
        """
        return [list(i) for i in self.km.cluster_centers_]

    def to_metadata(self):
        """Editing the metadata of the function
        """
        return {
            "k": self.k,
            "init": self.init,
            "verbose": self.verbose,
            "compute_labels": self.compute_labels,
            "max_no_improvement": self.max_no_improvement,
            "number_of_clusters": self.k
        }

# class KMedoids(CentroidCluster):
#     def fit_transform(self, 
#         vectors: np.ndarray, 
#         cluster_args: Optional[Dict[Any, Any]] = CLUSTER_DEFAULT_ARGS['kmedoids'], 
#         k: Union[None, int] = 10,
#     ) -> np.ndarray:
#         try:
#             from sklearn_extra.cluster import KMedoids
#         except ModuleNotFoundError as e:
#             raise ModuleNotFoundError(
#                 f"{e}\nInstall umap\n \
#                 pip install -U relevanceai[kmedoids]"
#             )
#         self.logger.debug(f"{cluster_args}")
#         km = KMedoids(n_clusters=k, **cluster_args).fit(vectors)
#         cluster_labels = km.labels_
#         # cluster_centroids = km.cluster_centers_
#         return cluster_labels

class KMeans(MiniBatchKMeans):
    def __init__(
        self, 
        k=10,
        init="k-means++",
        n_init=10,
        max_iter=300,
        tol=1e-4,
        verbose=0,
        random_state=None,
        copy_x=True,
        algorithm="auto", 
    ):
        self.init = init
        self.n_init = n_init
        self.max_iter = max_iter
        self.tol = tol
        self.verbose = verbose
        self.random_state = random_state
        self.copy_x = copy_x
        self.algorithm = algorithm
        self.n_clusters = k

    def _init_model(self):
        from sklearn.cluster import KMeans
        self.km = KMeans(
            n_clusters=self.n_clusters,
            init=self.init,
            verbose=self.verbose,
            max_iter=self.max_iter,
            tol=self.tol,
            random_state=self.random_state,
            copy_x=self.copy_x,
            algorithm=self.algorithm
        )
        return

    def to_metadata(self):
        """Editing the metadata of the function
        """
        return {
            "n_clusters":self.n_clusters,
            "init":self.init,
            "max_iter":self.max_iter,
            "tol":self.tol,
            "random_state":self.random_state,
            "copy_x":self.copy_x,
            "algorithm":self.algorithm,
        }

class HDBSCAN(DensityCluster):
    def fit_transform(self, 
        vectors: np.ndarray, 
        cluster_args: Optional[Dict[Any, Any]] = CLUSTER_DEFAULT_ARGS['hdbscan'], 
        min_cluster_size: Union[None, int] = 10,
    ) -> np.ndarray:
        try:
            from hdbscan import HDBSCAN
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"{e}\nInstall hdbscan\n \
                pip install -U relevanceai[hdbscan]"
            )
        self.logger.debug(f"{cluster_args}")
        hdbscan = HDBSCAN(min_cluster_size=min_cluster_size, **cluster_args).fit(vectors)
        cluster_labels = hdbscan.labels_
  
        return cluster_labels


class Cluster(BatchAPIClient, ClusterBase):

    def __init__(self, project, api_key):
        self.project = project
        super().__init__(project, api_key)

    @staticmethod
    def _choose_k(vectors: np.ndarray):
        """ "
        Choose k clusters
        """
        # Partitioning methods
        # if check_type(self.cluster, CLUSTER_NUMERIC):
        """
        Scaled_inertia = inertia(k)/inertia(k=1) + (a * K)
        where a is penalty factor of num_clusters
        """
        warnings.warn("This method is not implemented yet k=10")
        return 10


    @staticmethod
    def cluster(
        vectors: np.ndarray,
        cluster: Union[CLUSTER, ClusterBase],
        cluster_args: Union[None, dict],
        k: Union[None, int] = None,
    ) -> np.ndarray:
        """
        Cluster vectors
        """
        if isinstance(cluster, str):
            if cluster_args is None:
                cluster_args = CLUSTER_DEFAULT_ARGS[cluster]
            if cluster in ['kmeans', 'kmedoids']:
                if (k is None and cluster_args is None) \
                    or ("n_clusters" not in cluster_args.keys()):
                    k = Cluster._choose_k(vectors)
                if cluster == "kmeans":
                    return KMeans(**cluster_args).fit_transform(vectors=vectors)
                elif cluster == "kmedoids":
                    raise NotImplementedError
                    # return KMedioids().fit_transform(vectors=vectors, cluster_args=cluster_args)
            elif cluster == "hdbscan":
                return HDBSCAN().fit_transform(vectors=vectors, cluster_args=cluster_args)
            
        elif isinstance(cluster, ClusterBase):
            return cluster().fit_transform(vectors=vectors, cluster_args=cluster_args)
        raise ValueError("Not valid cluster input.")

    def kmeans_cluster(
        self,
        dataset_id: str,
        vector_fields: list,
        filters: List = [],
        k: Union[None, int] = 10,
        init: str = "k-means++",
        n_init: int = 10,
        max_iter: int = 300,
        tol: float = 1e-4,
        verbose: bool = True,
        random_state: Optional[int] = None,
        copy_x: bool = True,
        algorithm: str ="auto",
        alias: str = "default",
        cluster_field: str="_cluster_",
        update_documents_chunksize: int = 50,
        overwrite: bool = False
    ):
        """
        This function performs all the steps required for Kmeans clustering:
        1- Loads the data
        2- Clusters the data
        3- Updates the data with clustering info
        4- Adds the centroid to the hidden centroid collection

        Parameters
        ----------
        dataset_id : string
            name of the dataser
        vector_fields : list
            a list containing the vector field to be used for clustering
        filters : list
            a list to filter documents of the dataset,
        k : int
            K in Kmeans
        init : string
            "k-means++" -> Kmeans algorithm parameter
        n_init : int
            number of reinitialization for the kmeans algorithm
        max_iter : int
            max iteration in the kmeans algorithm
        tol : int
            tol in the kmeans algorithm
        verbose : bool
            True by default
        random_state = None
            None by default -> Kmeans algorithm parameter
        copy_x : bool
            True bydefault
        algorithm : string
            "auto" by default
        alias : string
            "default", string to be used in naming of the field showing the clustering results
        cluster_field: string
            "_cluster_", string to name the main cluster field
        overwrite : bool
            False by default, To overwite an existing clusering result
        
        Example
        -------------

        >>> client.vector_tools.cluster.kmeans_cluster(
            dataset_id="sample_dataset",
            vector_fields=["sample_1_vector_"] # Only 1 vector field is supported for now
        )
        """

        EXISTING_CLUSTER_MESSAGE = """Clustering results already exist"""

        if '.'.join([cluster_field, vector_fields[0], alias]) in self.datasets.schema(dataset_id) and not overwrite:
            raise ClusteringResultsAlreadyExistsError(EXISTING_CLUSTER_MESSAGE)

        # load the documents
        docs = self.get_all_documents(dataset_id=dataset_id, filters=filters, select_fields=vector_fields)

        # Cluster
        clusterer = KMeans(
            k=k,
            init=init,
            n_init=n_init,
            max_iter=max_iter,
            tol=tol,
            verbose=verbose,
            random_state=random_state,
            copy_x=copy_x,
            algorithm=algorithm
        )
        clustered_docs = clusterer.fit_documents(
            vector_fields,
            docs,
            alias=alias, 
            cluster_field=cluster_field, 
            return_only_clusters=True)

        # Updating the db
        try:
            results = self.update_documents(dataset_id, clustered_docs, chunksize = update_documents_chunksize)
        except Exception as e:
            self.logger.error(e)
        self.logger.info(results)

        # Update the centroid collection
        centers = clusterer.get_centroid_docs()
        try:
            results = self.services.cluster.centroids.insert(
                dataset_id = dataset_id,
                cluster_centers=centers,
                vector_field=vector_fields[0],
                alias= alias
            )
        except Exception as e:
            self.logger.error(e)
        self.logger.info(results)
        return centers



"""Missing field error
"""
class RelevanceAIError(Exception):
    """base class for all errors
    """

class MissingFieldError(RelevanceAIError):
    """Error handling for missing fields"""


class APIError(RelevanceAIError):
    """Error related to API"""

class ClusteringResultsAlreadyExistsError(RelevanceAIError):
    """Error is raised when the clustering dataset already exists
    """
    pass



"""Get a good progress bar
"""
# from contextlib import nullcontext
from contextlib import AbstractContextManager


class ProgressBar:
    def __call__(self, iterable):
        return self.get_bar()(iterable)

    @staticmethod
    def is_in_ipython():
        """
        Determines if current code is executed within an ipython session.
        """
        is_in_ipython = False
        # Check if the runtime is within an interactive environment, i.e., ipython.
        try:
            from IPython import get_ipython  # pylint: disable=import-error

            if get_ipython():
                is_in_ipython = True
        except ImportError:
            pass  # If dependencies are not available, then not interactive for sure.
        return is_in_ipython

    def is_in_notebook(self) -> bool:
        """
        Determines if current code is executed from an ipython notebook.
        """
        is_in_notebook = False
        if self.is_in_ipython():
            # The import and usage must be valid under the execution path.
            from IPython import get_ipython

            if "IPKernelApp" in get_ipython().config:
                is_in_notebook = True
        return is_in_notebook

    def get_bar(self):
        from tqdm.auto import tqdm

        return tqdm


class NullProgressBar(AbstractContextManager):
    """Context manager that does no additional processing.

    Used as a stand-in for a normal context manager, when a particular
    block of code is only sometimes used with a normal context manager:

    cm = optional_cm if condition else nullcontext()
    with cm:
        # Perform operation, using optional_cm if condition is True
    """

    def __init__(self, iterable: int = None):
        self.iterable = iterable

    def __enter__(self):
        return self.iterable

    def __exit__(self, *excinfo):
        pass

    def __iter__(self):
        if self.iterable is None:
            self.iterable = range(0)
        for i in self.iterable:
            yield i


def progress_bar(iterable, show_progress_bar: bool = False):

    try:
        if show_progress_bar:
            return ProgressBar()(iterable)
    except Exception as e:
        pass
    return NullProgressBar(iterable)




"""Batch client to allow for batch insertions/retrieval and encoding
"""
from typing import Callable

from relevanceai.api.batch.batch_insert import BatchInsert
from relevanceai.api.batch.batch_retrieve import BatchRetrieve
from relevanceai.api.endpoints.client import APIClient
from relevanceai.api.batch.chunk import Chunker


class BatchAPIClient(BatchInsert, BatchRetrieve, APIClient):
    """Batch API client"""

    def batch_insert(self):
        raise NotImplemented

    def batch_get_and_edit(self, dataset_id: str, chunk_size: int, bulk_edit: Callable):
        """Batch get the documents and return the documents"""
        raise NotImplemented



"""Recommmend services.
"""
from relevanceai.base import Base


class Recommend(Base):
    def __init__(self, project, api_key):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)

    def vector(
        self,
        dataset_id: str,
        positive_document_ids: dict = {},
        negative_document_ids: dict = {},
        vector_fields=[],
        approximation_depth: int = 0,
        vector_operation: str = "sum",
        sum_fields: bool = True,
        page_size: int = 20,
        page: int = 1,
        similarity_metric: str = "cosine",
        facets: list = [],
        filters: list = [],
        min_score: float = 0,
        select_fields: list = [],
        include_vector: bool = False,
        include_count: bool = True,
        asc: bool = False,
        keep_search_history: bool = False,
        hundred_scale: bool = False,
    ):

        """ 
        Vector Search based recommendations are done by extracting the vectors of the documents ids specified performing some vector operations and then searching the dataset with the resultant vector. This allows us to not only do recommendations but personalized and weighted recommendations. \n
        Here are a couple of different scenarios and what the queries would look like for those: \n

        Recommendations Personalized by single liked product:

        >>> positive_document_ids=['A']

        -> Document ID A Vector = Search Query \n

        Recommendations Personalized by multiple liked product:

        >>> positive_document_ids=['A', 'B']

        -> Document ID A Vector + Document ID B Vector = Search Query \n

        Recommendations Personalized by multiple liked product and disliked products:

        >>> positive_document_ids=['A', 'B'], negative_document_ids=['C', 'D']

        -> (Document ID A Vector + Document ID B Vector) - (Document ID C Vector + Document ID C Vector) = Search Query \n

        Recommendations Personalized by multiple liked product and disliked products with weights:

        >>> positive_document_ids={'A':0.5, 'B':1}, negative_document_ids={'C':0.6, 'D':0.4}

        -> (Document ID A Vector * 0.5 + Document ID B Vector * 1) - (Document ID C Vector * 0.6 + Document ID D Vector * 0.4) = Search Query \n

        You can change the operator between vectors with vector_operation:

        e.g. >>> positive_document_ids=['A', 'B'], negative_document_ids=['C', 'D'], vector_operation='multiply'

        -> (Document ID A Vector * Document ID B Vector) - (Document ID C Vector * Document ID D Vector) = Search Query

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        positive_document_ids : dict
            Positive document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        negative_document_ids: dict
            Negative document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        vector_fields: list
            The vector field to search in. It can either be an array of strings (automatically equally weighted) (e.g. ['check_vector_', 'yellow_vector_']) or it is a dictionary mapping field to float where the weighting is explicitly specified (e.g. {'check_vector_': 0.2, 'yellow_vector_': 0.5})
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        vector_operation: string
            Aggregation for the vectors when using positive and negative document IDs, choose from ['mean', 'sum', 'min', 'max', 'divide', 'mulitple']
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into VecDB. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        """

        return self.make_http_request(
            f"/services/recommend/vector",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "positive_document_ids": positive_document_ids,
                "negative_document_ids": negative_document_ids,
                "vector_fields": vector_fields,
                "approximation_depth": approximation_depth,
                "vector_operation": vector_operation,
                "sum_fields": sum_fields,
                "page_size": page_size,
                "page": page,
                "similarity_metric": similarity_metric,
                "facets": facets,
                "filters": filters,
                "min_score": min_score,
                "select_fields": select_fields,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
                "keep_search_history": keep_search_history,
                "hundred_scale": hundred_scale
            }
        )

    def diversity(
        self,
        dataset_id: str,
        cluster_vector_field: str,
        n_clusters: int,
        positive_document_ids: dict = {},
        negative_document_ids: dict = {},
        vector_fields=[],
        approximation_depth: int = 0,
        vector_operation: str = "sum",
        sum_fields: bool = True,
        page_size: int = 20,
        page: int = 1,
        similarity_metric: str = "cosine",
        facets: list = [],
        filters: list = [],
        min_score: float = 0,
        select_fields: list = [],
        include_vector: bool = False,
        include_count: bool = True,
        asc: bool = False,
        keep_search_history: bool = False,
        hundred_scale: bool = False,
        search_history_id: str = None,
        n_init: int = 5,
        n_iter: int = 10,
        return_as_clusters: bool = False
    ):

        """ 
        Vector Search based recommendations are done by extracting the vectors of the documents ids specified performing some vector operations and then searching the dataset with the resultant vector. This allows us to not only do recommendations but personalized and weighted recommendations. \n
        Diversity recommendation increases the variety within the recommendations via clustering. Search results are clustered and the top k items in each cluster are selected. The main clustering parameters are cluster_vector_field and n_clusters, the vector field on which to perform clustering and number of clusters respectively. \n
        Here are a couple of different scenarios and what the queries would look like for those: \n

        Recommendations Personalized by single liked product:

        >>> positive_document_ids=['A']

        -> Document ID A Vector = Search Query

        Recommendations Personalized by multiple liked product:

        >>> positive_document_ids=['A', 'B']

        -> Document ID A Vector + Document ID B Vector = Search Query

        Recommendations Personalized by multiple liked product and disliked products:

        >>> positive_document_ids=['A', 'B'], negative_document_ids=['C', 'D']

        -> (Document ID A Vector + Document ID B Vector) - (Document ID C Vector + Document ID C Vector) = Search Query

        Recommendations Personalized by multiple liked product and disliked products with weights:

        >>> positive_document_ids={'A':0.5, 'B':1}, negative_document_ids={'C':0.6, 'D':0.4}

        -> (Document ID A Vector * 0.5 + Document ID B Vector * 1) - (Document ID C Vector * 0.6 + Document ID D Vector * 0.4) = Search Query

        You can change the operator between vectors with vector_operation:

        e.g. >>> positive_document_ids=['A', 'B'], negative_document_ids=['C', 'D'], vector_operation='multiply'

        -> (Document ID A Vector * Document ID B Vector) - (Document ID C Vector * Document ID D Vector) = Search Query

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        cluster_vector_field: str
            The field to cluster on.
        n_clusters: int
            Number of clusters to be specified.
        positive_document_ids : dict
            Positive document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        negative_document_ids: dict
            Negative document IDs to personalize the results with, this will retrive the vectors from the document IDs and consider it in the operation.
        vector_fields: list
            The vector field to search in. It can either be an array of strings (automatically equally weighted) (e.g. ['check_vector_', 'yellow_vector_']) or it is a dictionary mapping field to float where the weighting is explicitly specified (e.g. {'check_vector_': 0.2, 'yellow_vector_': 0.5})
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        vector_operation: string
            Aggregation for the vectors when using positive and negative document IDs, choose from ['mean', 'sum', 'min', 'max', 'divide', 'mulitple']
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        facets: list
            Fields to include in the facets, if [] then all
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        select_fields: list
            Fields to include in the search results, empty array/list means all fields.
        include_vector: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        asc: bool
            Whether to sort results by ascending or descending order
        keep_search_history: bool
            Whether to store the history into VecDB. This will increase the storage costs over time.
        hundred_scale: bool
            Whether to scale up the metric by 100
        search_history_id: str
            Search history ID, only used for storing search histories.
        n_init: int
            Number of runs to run with different centroid seeds
        n_iter: int
            Number of iterations in each run
        return_as_clusters: bool
            If True, return as clusters as opposed to results list
        """

        return self.make_http_request(
            f"/services/recommend/diversity",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "positive_document_ids": positive_document_ids,
                "negative_document_ids": negative_document_ids,
                "vector_fields": vector_fields,
                "approximation_depth": approximation_depth,
                "vector_operation": vector_operation,
                "sum_fields": sum_fields,
                "page_size": page_size,
                "page": page,
                "similarity_metric": similarity_metric,
                "facets": facets,
                "filters": filters,
                "min_score": min_score,
                "select_fields": select_fields,
                "include_vector": include_vector,
                "include_count": include_count,
                "asc": asc,
                "keep_search_history": keep_search_history,
                "hundred_scale": hundred_scale,
                "search_history_id": search_history_id,
                "cluster_vector_field": cluster_vector_field,
                "n_clusters": n_clusters,
                "n_init": n_init,
                "n_iter": n_iter,
                "return_as_clusters": return_as_clusters
            }
        )




"""Services class
"""
from relevanceai.base import Base

from relevanceai.api.endpoints.encoders import Encoders
from relevanceai.api.endpoints.cluster import Cluster
from relevanceai.api.endpoints.search import Search
from relevanceai.api.endpoints.aggregate import Aggregate
from relevanceai.api.endpoints.recommend import Recommend
from relevanceai.api.endpoints.tagger import Tagger
from relevanceai.api.endpoints.prediction import Prediction
from relevanceai.api.endpoints.wordclouds import Wordclouds

class Services(Base):
    def __init__(self, project: str, api_key: str):
        self.project = project
        self.api_key = api_key
        self.encoders = Encoders(project=project, api_key=api_key)
        self.cluster = Cluster(project=project, api_key=api_key)
        self.search = Search(project=project, api_key=api_key)
        self.aggregate = Aggregate(project=project, api_key=api_key)
        self.recommend = Recommend(project=project, api_key=api_key)
        self.tagger = Tagger(project=project, api_key=api_key)
        self.prediction = Prediction(project=project, api_key=api_key)
        self.wordclouds = Wordclouds(project=project, api_key=api_key)
        super().__init__(project, api_key)

    def document_diff(self, doc: dict, docs_to_compare: list, difference_fields: list = []):
        """ 
        Find differences between documents

        Parameters
        ----------
        doc: dict
            Main document to compare other documents against.
        docs_to_compare: list
            Other documents to compare against the main document.
        difference_fields: list
            Fields to compare. Defaults to [], which compares all fields.
    
        """
        return self.make_http_request(
            endpoint=f"/services/document_diff",
            method="POST",
            parameters={
                "doc": doc,
                "docs_to_compare": docs_to_compare,
                "difference_fields": difference_fields,
            }
        )

